@@ -7,8 +7,8 @@
    This file is part of Massif, a Valgrind tool for profiling memory
    usage of programs.
 
-   Copyright (C) 2003-2004 Nicholas Nethercote
-      njn25@cam.ac.uk
+   Copyright (C) 2003-2005 Nicholas Nethercote
+      njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -34,8 +34,22 @@
 // be inspected step-wise to an appropriate depth.  See comments on data
 // structures below for more info on how things work.
 
-#include "vg_skin.h"
-//#include "vg_profile.c"
+#include "pub_tool_basics.h"
+#include "pub_tool_debuginfo.h"
+#include "pub_tool_hashtable.h"
+#include "pub_tool_libcbase.h"
+#include "pub_tool_libcassert.h"
+#include "pub_tool_libcfile.h"
+#include "pub_tool_libcmman.h"
+#include "pub_tool_libcprint.h"
+#include "pub_tool_libcproc.h"
+#include "pub_tool_machine.h"
+#include "pub_tool_mallocfree.h"
+#include "pub_tool_options.h"
+#include "pub_tool_profile.h"
+#include "pub_tool_replacemalloc.h"
+#include "pub_tool_stacktrace.h"
+#include "pub_tool_tooliface.h"
 
 #include "valgrind.h"           // For {MALLOC,FREE}LIKE_BLOCK
 
@@ -90,7 +104,7 @@
 typedef struct _XPt XPt;
 
 struct _XPt {
-   Addr  eip;              // code address
+   Addr  ip;              // code address
 
    // Bottom-XPts: space for the precise context.
    // Other XPts:  space of all the descendent bottom-XPts.
@@ -178,7 +192,7 @@ typedef
    struct _HP_Chunk {
       struct _HP_Chunk* next;
       Addr              data;    // Ptr to actual block
-      UInt              size;    // Size requested
+      SizeT             size;    // Size requested
       XPt*              where;   // Where allocated; bottom-XPt
    }
    HP_Chunk;
@@ -200,7 +214,7 @@ typedef
       VgpPrintHp,
       VgpPrintXPts,
    }
-   VgpSkinCC;
+   VgpToolCC;
 
 /*------------------------------------------------------------*/
 /*--- Statistics                                           ---*/
@@ -244,12 +258,14 @@ static Char buf [BUF_LEN];
 static Char buf2[BUF_LEN];
 static Char buf3[BUF_LEN];
 
-static UInt sigstacks_space = 0;    // Current signal stacks space sum
+static SizeT sigstacks_space = 0;    // Current signal stacks space sum
 
 static VgHashTable malloc_list  = NULL;   // HP_Chunks
 
 static UInt n_heap_blocks = 0;
 
+// Current directory at startup.
+static Char base_dir[VKI_PATH_MAX];
 
 #define MAX_ALLOC_FNS      32      // includes the builtin ones
 
@@ -265,7 +281,6 @@ static Char* alloc_fns[MAX_ALLOC_FNS] = {
    "__builtin_vec_new",
    "calloc",
    "realloc",
-   "my_malloc",   // from vg_libpthread.c
    "memalign",
 };
 
@@ -288,13 +303,13 @@ static Bool clo_stacks      = True;
 static Bool clo_depth       = 3;
 static XFormat clo_format   = XText;
 
-Bool SK_(process_cmd_line_option)(Char* arg)
+static Bool ms_process_cmd_line_option(Char* arg)
 {
-        VG_BOOL_CLO("--heap",       clo_heap)
-   else VG_BOOL_CLO("--stacks",     clo_stacks)
+        VG_BOOL_CLO(arg, "--heap",       clo_heap)
+   else VG_BOOL_CLO(arg, "--stacks",     clo_stacks)
 
-   else VG_NUM_CLO ("--heap-admin",  clo_heap_admin)
-   else VG_BNUM_CLO("--depth",       clo_depth, 1, MAX_DEPTH)
+   else VG_NUM_CLO (arg, "--heap-admin",  clo_heap_admin)
+   else VG_BNUM_CLO(arg, "--depth",       clo_depth, 1, MAX_DEPTH)
 
    else if (VG_CLO_STREQN(11, arg, "--alloc-fn=")) {
       alloc_fns[n_alloc_fns] = & arg[11];
@@ -316,7 +331,7 @@ Bool SK_(process_cmd_line_option)(Char* arg)
    return True;
 }
 
-void SK_(print_usage)(void)
+static void ms_print_usage(void)
 {
    VG_(printf)( 
 "    --heap=no|yes             profile heap blocks [yes]\n"
@@ -329,7 +344,7 @@ void SK_(print_usage)(void)
    VG_(replacement_malloc_print_usage)();
 }
 
-void SK_(print_debug_usage)(void)
+static void ms_print_debug_usage(void)
 {
    VG_(replacement_malloc_print_debug_usage)();
 }
@@ -344,7 +359,7 @@ static XPt* alloc_xpt;
 
 // Cheap allocation for blocks that never need to be freed.  Saves about 10%
 // for Konqueror startup with --depth=40.
-static void* perm_malloc(UInt n_bytes)
+static void* perm_malloc(SizeT n_bytes)
 {
    static Addr hp     = 0;    // current heap pointer
    static Addr hp_lim = 0;    // maximum usable byte in current block
@@ -363,10 +378,10 @@ static void* perm_malloc(UInt n_bytes)
 
 
 
-static XPt* new_XPt(Addr eip, XPt* parent, Bool is_bottom)
+static XPt* new_XPt(Addr ip, XPt* parent, Bool is_bottom)
 {
    XPt* xpt          = perm_malloc(sizeof(XPt));
-   xpt->eip          = eip;
+   xpt->ip           = ip;
 
    xpt->curr_space    = 0;
    xpt->approx_ST     = 0;
@@ -375,7 +390,7 @@ static XPt* new_XPt(Addr eip, XPt* parent, Bool is_bottom)
    xpt->parent       = parent;
 
    // Check parent is not a bottom-XPt
-   sk_assert(parent == NULL || 0 != parent->max_children);
+   tl_assert(parent == NULL || 0 != parent->max_children);
 
    xpt->n_children   = 0;
 
@@ -397,11 +412,11 @@ static XPt* new_XPt(Addr eip, XPt* parent, Bool is_bottom)
    return xpt;
 }
 
-static Bool is_alloc_fn(Addr eip)
+static Bool is_alloc_fn(Addr ip)
 {
    Int i;
 
-   if ( VG_(get_fnname)(eip, buf, BUF_LEN) ) {
+   if ( VG_(get_fnname)(ip, buf, BUF_LEN) ) {
       for (i = 0; i < n_alloc_fns; i++) {
          if (VG_STREQ(buf, alloc_fns[i]))
             return True;
@@ -415,11 +430,11 @@ static Bool is_alloc_fn(Addr eip)
 // to ensure this in certain cases.  See comments below.
 static XPt* get_XCon( ThreadId tid, Bool custom_malloc )
 {
-   // Static to minimise stack size.  +1 for added 0xffffffff %eip.
-   static Addr eips[MAX_DEPTH + MAX_ALLOC_FNS + 1];
+   // Static to minimise stack size.  +1 for added ~0 IP
+   static Addr ips[MAX_DEPTH + MAX_ALLOC_FNS + 1];
 
    XPt* xpt = alloc_xpt;
-   UInt n_eips, L, A, B, nC;
+   UInt n_ips, L, A, B, nC;
    UInt overestimate;
    Bool reached_bottom;
 
@@ -432,42 +447,42 @@ static XPt* get_XCon( ThreadId tid, Bool custom_malloc )
    // it is enough.
    overestimate = 2;
    while (True) {
-      n_eips = VG_(stack_snapshot)( tid, eips, clo_depth + overestimate );
+      n_ips = VG_(get_StackTrace)( tid, ips, clo_depth + overestimate );
 
-      // Now we add a dummy "unknown" %eip at the end.  This is only used if we
-      // run out of %eips before hitting clo_depth.  It's done to ensure the
+      // Now we add a dummy "unknown" IP at the end.  This is only used if we
+      // run out of IPs before hitting clo_depth.  It's done to ensure the
       // XPt we return is (now and forever) a bottom-XPt.  If the returned XPt
       // wasn't a bottom-XPt (now or later) it would cause problems later (eg.
       // the parent's approx_ST wouldn't be equal [or almost equal] to the
       // total of the childrens' approx_STs).  
-      eips[ n_eips++ ] = 0xffffffff;
+      ips[ n_ips++ ] = ~((Addr)0);
 
-      // Skip over alloc functions in eips[]. 
-      for (L = 0; is_alloc_fn(eips[L]) && L < n_eips; L++) { }
+      // Skip over alloc functions in ips[]. 
+      for (L = 0; is_alloc_fn(ips[L]) && L < n_ips; L++) { }
 
       // Must be at least one alloc function, unless client used
       // MALLOCLIKE_BLOCK
-      if (!custom_malloc) sk_assert(L > 0);    
+      if (!custom_malloc) tl_assert(L > 0);    
 
       // Should be at least one non-alloc function.  If not, try again.
-      if (L == n_eips) {
+      if (L == n_ips) {
          overestimate += 2;
          if (overestimate > MAX_ALLOC_FNS)
-            VG_(skin_panic)("No stk snapshot big enough to find non-alloc fns");
+            VG_(tool_panic)("No stk snapshot big enough to find non-alloc fns");
       } else {
          break;
       }
    }
    A = L;
-   B = n_eips - 1;
+   B = n_ips - 1;
    reached_bottom = False;
 
-   // By this point, the eips we care about are in eips[A]..eips[B]
+   // By this point, the IPs we care about are in ips[A]..ips[B]
 
    // Now do the search/insertion of the XCon. 'L' is the loop counter,
-   // being the index into eips[].
+   // being the index into ips[].
    while (True) {
-      // Look for %eip in xpt's children.
+      // Look for IP in xpt's children.
       // XXX: linear search, ugh -- about 10% of time for konqueror startup
       // XXX: tried cacheing last result, only hit about 4% for konqueror
       // Nb:  this search hits about 98% of the time for konqueror
@@ -482,8 +497,8 @@ static XPt* get_XCon( ThreadId tid, Bool custom_malloc )
       while (True) {
          if (nC == xpt->n_children) {
             // not found, insert new XPt
-            sk_assert(xpt->max_children != 0);
-            sk_assert(xpt->n_children <= xpt->max_children);
+            tl_assert(xpt->max_children != 0);
+            tl_assert(xpt->n_children <= xpt->max_children);
             // Expand 'children' if necessary
             if (xpt->n_children == xpt->max_children) {
                xpt->max_children *= 2;
@@ -491,19 +506,19 @@ static XPt* get_XCon( ThreadId tid, Bool custom_malloc )
                                              xpt->max_children * sizeof(XPt*) );
                n_children_reallocs++;
             }
-            // Make new XPt for %eip, insert in list
+            // Make new XPt for IP, insert in list
             xpt->children[ xpt->n_children++ ] = 
-               new_XPt(eips[L], xpt, reached_bottom);
+               new_XPt(ips[L], xpt, reached_bottom);
             break;
          }
-         if (eips[L] == xpt->children[nC]->eip) break;   // found the %eip
+         if (ips[L] == xpt->children[nC]->ip) break;   // found the IP
          nC++;                                           // keep looking
       }
       VGP_POPCC(VgpGetXPtSearch);
 
       // Return found/built bottom-XPt.
       if (reached_bottom) {
-         sk_assert(0 == xpt->children[nC]->n_children);   // Must be bottom-XPt
+         tl_assert(0 == xpt->children[nC]->n_children);   // Must be bottom-XPt
          VGP_POPCC(VgpGetXPt);
          return xpt->children[nC];
       }
@@ -519,17 +534,17 @@ static void update_XCon(XPt* xpt, Int space_delta)
 {
    VGP_PUSHCC(VgpUpdateXCon);
 
-   sk_assert(True == clo_heap);
-   sk_assert(0    != space_delta);
-   sk_assert(NULL != xpt);
-   sk_assert(0    == xpt->n_children);    // must be bottom-XPt
+   tl_assert(True == clo_heap);
+   tl_assert(0    != space_delta);
+   tl_assert(NULL != xpt);
+   tl_assert(0    == xpt->n_children);    // must be bottom-XPt
 
    while (xpt != alloc_xpt) {
-      if (space_delta < 0) sk_assert(xpt->curr_space >= -space_delta);
+      if (space_delta < 0) tl_assert(xpt->curr_space >= -space_delta);
       xpt->curr_space += space_delta;
       xpt = xpt->parent;
    } 
-   if (space_delta < 0) sk_assert(alloc_xpt->curr_space >= -space_delta);
+   if (space_delta < 0) tl_assert(alloc_xpt->curr_space >= -space_delta);
    alloc_xpt->curr_space += space_delta;
 
    VGP_POPCC(VgpUpdateXCon);
@@ -602,7 +617,7 @@ static void adjust(Queue* q)
 {
    void** old_elems;
 
-   sk_assert(q->tail == q->max_elems);
+   tl_assert(q->tail == q->max_elems);
    if (q->head < 10) {
       old_elems     = q->elems;
       q->max_elems *= 2; 
@@ -648,14 +663,14 @@ void add_HP_Chunk(HP_Chunk* hc)
 static __inline__ 
 HP_Chunk* get_HP_Chunk(void* p, HP_Chunk*** prev_chunks_next_ptr)
 {
-   return (HP_Chunk*)VG_(HT_get_node) ( malloc_list, (UInt)p,
+   return (HP_Chunk*)VG_(HT_get_node) ( malloc_list, (UWord)p,
                                         (VgHashNode***)prev_chunks_next_ptr );
 }
 
 static __inline__
 void remove_HP_Chunk(HP_Chunk* hc, HP_Chunk** prev_chunks_next_ptr)
 {
-   sk_assert(n_heap_blocks > 0);
+   tl_assert(n_heap_blocks > 0);
    n_heap_blocks--;
    *prev_chunks_next_ptr = hc->next;
 }
@@ -663,8 +678,9 @@ void remove_HP_Chunk(HP_Chunk* hc, HP_Chunk** prev_chunks_next_ptr)
 // Forward declaration
 static void hp_census(void);
 
-static __inline__
-void* new_block ( void* p, Int size, UInt align, Bool is_zeroed )
+static
+void* new_block ( ThreadId tid, void* p, SizeT size, SizeT align,
+                  Bool is_zeroed )
 {
    HP_Chunk* hc;
    Bool custom_alloc = (NULL == p);
@@ -692,7 +708,7 @@ void* new_block ( void* p, Int size, UInt align, Bool is_zeroed )
    hc->data = (Addr)p;
    hc->where = NULL;    // paranoia
    if (clo_heap) {
-      hc->where = get_XCon( VG_(get_current_or_recent_tid)(), custom_alloc );
+      hc->where = get_XCon( tid, custom_alloc );
       if (0 != size) 
          update_XCon(hc->where, size);
    }
@@ -719,7 +735,7 @@ void die_block ( void* p, Bool custom_free )
    hc = get_HP_Chunk( p, &remove_handle );
    if (hc == NULL)
       return;   // must have been a bogus free(), or p==NULL
-   sk_assert(hc->data == (Addr)p);
+   tl_assert(hc->data == (Addr)p);
    remove_HP_Chunk(hc, remove_handle);
 
    if (clo_heap && hc->size != 0)
@@ -738,53 +754,53 @@ void die_block ( void* p, Bool custom_free )
 }
  
 
-void* SK_(malloc) ( Int n )
+static void* ms_malloc ( ThreadId tid, SizeT n )
 {
-   return new_block( NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
+   return new_block( tid, NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
 }
 
-void* SK_(__builtin_new) ( Int n )
+static void* ms___builtin_new ( ThreadId tid, SizeT n )
 {
-   return new_block( NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
+   return new_block( tid, NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
 }
 
-void* SK_(__builtin_vec_new) ( Int n )
+static void* ms___builtin_vec_new ( ThreadId tid, SizeT n )
 {
-   return new_block( NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
+   return new_block( tid, NULL, n, VG_(clo_alignment), /*is_zeroed*/False );
 }
 
-void* SK_(calloc) ( Int m, Int size )
+static void* ms_calloc ( ThreadId tid, SizeT m, SizeT size )
 {
-   return new_block( NULL, m*size, VG_(clo_alignment), /*is_zeroed*/True );
+   return new_block( tid, NULL, m*size, VG_(clo_alignment), /*is_zeroed*/True );
 }
 
-void *SK_(memalign)( Int align, Int n )
+static void *ms_memalign ( ThreadId tid, SizeT align, SizeT n )
 {
-   return new_block( NULL, n, align, False );
+   return new_block( tid, NULL, n, align, False );
 }
 
-void SK_(free) ( void* p )
+static void ms_free ( ThreadId tid, void* p )
 {
    die_block( p, /*custom_free*/False );
 }
 
-void SK_(__builtin_delete) ( void* p )
+static void ms___builtin_delete ( ThreadId tid, void* p )
 {
    die_block( p, /*custom_free*/False);
 }
 
-void SK_(__builtin_vec_delete) ( void* p )
+static void ms___builtin_vec_delete ( ThreadId tid, void* p )
 {
    die_block( p, /*custom_free*/False );
 }
 
-void* SK_(realloc) ( void* p_old, Int new_size )
+static void* ms_realloc ( ThreadId tid, void* p_old, SizeT new_size )
 {
    HP_Chunk*    hc;
    HP_Chunk**   remove_handle;
    Int          i;
    void*        p_new;
-   UInt         old_size;
+   SizeT        old_size;
    XPt         *old_where, *new_where;
    
    VGP_PUSHCC(VgpCliMalloc);
@@ -796,7 +812,7 @@ void* SK_(realloc) ( void* p_old, Int new_size )
       return NULL;   // must have been a bogus free()
    }
 
-   sk_assert(hc->data == (Addr)p_old);
+   tl_assert(hc->data == (Addr)p_old);
    old_size = hc->size;
   
    if (new_size <= old_size) {
@@ -814,8 +830,7 @@ void* SK_(realloc) ( void* p_old, Int new_size )
    }
    
    old_where = hc->where;
-   new_where = get_XCon( VG_(get_current_or_recent_tid)(),
-                         /*custom_malloc*/False);
+   new_where = get_XCon( tid, /*custom_malloc*/False);
 
    // Update HP_Chunk
    hc->data  = (Addr)p_new;
@@ -917,7 +932,7 @@ static void halve_censi(void)
 
    // Sets j to the index of the first not-yet-removed census at or after i
    #define FIND_CENSUS(i, j) \
-      for (j = i; -1 == censi[j].ms_time; j++) { }
+      for (j = i; j < MAX_N_CENSI && -1 == censi[j].ms_time; j++) { }
 
    for (i = 2; i < MAX_N_CENSI; i += 2) {
       // Find the censi representing the smallest timespan.  The timespan
@@ -933,7 +948,7 @@ static void halve_censi(void)
       FIND_CENSUS(j+1, jn);
       while (jn < MAX_N_CENSI) {
          Int timespan = censi[jn].ms_time - censi[jp].ms_time;
-         sk_assert(timespan >= 0);
+         tl_assert(timespan >= 0);
          if (timespan < min_span) {
             min_span = timespan;
             min_j    = j;
@@ -983,7 +998,6 @@ static void hp_census(void)
    static UInt ms_next_census = 0;     // zero allows startup census
 
    Int     ms_time, ms_time_since_prev;
-   Int     i, K;
    Census* census;
 
    VGP_PUSHCC(VgpCensus);
@@ -1004,6 +1018,7 @@ static void hp_census(void)
 
    // Heap: snapshot the K most significant XTrees -------------------
    if (clo_heap) {
+      Int i, K;
       K = ( alloc_xpt->n_children < MAX_SNAPSHOTS 
           ? alloc_xpt->n_children
           : MAX_SNAPSHOTS);     // max out
@@ -1053,7 +1068,7 @@ static void hp_census(void)
          VGP_PUSHCC(VgpCensusSnapshot);
          xtree_size2 = do_space_snapshot(alloc_xpt->children[i],
                                          census->xtree_snapshots[i], 0);
-         sk_assert(xtree_size == xtree_size2);
+         tl_assert(xtree_size == xtree_size2);
          VGP_POPCC(VgpCensusSnapshot);
       }
 //      VG_(printf)("\n\n");
@@ -1080,7 +1095,6 @@ static void hp_census(void)
       census->stacks_space = sigstacks_space;
       // slightly abusing this function
       VG_(first_matching_thread_stack)( count_stack_size, &census->stacks_space );
-      i++;
    }
 
    // Finish, update interval if necessary -----------------------------
@@ -1113,14 +1127,14 @@ static void hp_census(void)
 /*--- Tracked events                                       ---*/
 /*------------------------------------------------------------*/
 
-static void new_mem_stack_signal(Addr a, UInt len)
+static void new_mem_stack_signal(Addr a, SizeT len)
 {
    sigstacks_space += len;
 }
 
-static void die_mem_stack_signal(Addr a, UInt len)
+static void die_mem_stack_signal(Addr a, SizeT len)
 {
-   sk_assert(sigstacks_space >= len);
+   tl_assert(sigstacks_space >= len);
    sigstacks_space -= len;
 }
 
@@ -1128,16 +1142,16 @@ static void die_mem_stack_signal(Addr a, UInt len)
 /*--- Client Requests                                      ---*/
 /*------------------------------------------------------------*/
 
-Bool SK_(handle_client_request) ( ThreadId tid, UInt* argv, UInt* ret )
+static Bool ms_handle_client_request ( ThreadId tid, UWord* argv, UWord* ret )
 {
    switch (argv[0]) {
    case VG_USERREQ__MALLOCLIKE_BLOCK: {
       void* res;
       void* p         = (void*)argv[1];
-      UInt  sizeB     =        argv[2];
+      SizeT sizeB     =        argv[2];
       *ret            = 0;
-      res = new_block( p, sizeB, /*align -- ignored*/0, /*is_zeroed*/False );
-      sk_assert(res == p);
+      res = new_block( tid, p, sizeB, /*align--ignored*/0, /*is_zeroed*/False );
+      tl_assert(res == p);
       return True;
    }
    case VG_USERREQ__FREELIKE_BLOCK: {
@@ -1152,68 +1166,15 @@ Bool SK_(handle_client_request) ( ThreadId tid, UInt* argv, UInt* ret )
    }
 }
 
-/*------------------------------------------------------------*/
-/*--- Initialisation                                       ---*/
-/*------------------------------------------------------------*/
-
-// Current directory at startup.
-static Char* base_dir;
-
-UInt VG_(vg_malloc_redzone_szB) = 0;
-
-void SK_(pre_clo_init)()
-{ 
-   VG_(details_name)            ("Massif");
-   VG_(details_version)         (NULL);
-   VG_(details_description)     ("a space profiler");
-   VG_(details_copyright_author)("Copyright (C) 2003, Nicholas Nethercote");
-   VG_(details_bug_reports_to)  (VG_BUGS_TO);
-
-   // Needs
-   VG_(needs_libc_freeres)();
-   VG_(needs_command_line_options)();
-   VG_(needs_client_requests)     ();
-
-   // Events to track
-   VG_(init_new_mem_stack_signal) ( new_mem_stack_signal );
-   VG_(init_die_mem_stack_signal) ( die_mem_stack_signal );
-
-   // Profiling events
-   VGP_(register_profile_event)(VgpGetXPt,         "get-XPt");
-   VGP_(register_profile_event)(VgpGetXPtSearch,   "get-XPt-search");
-   VGP_(register_profile_event)(VgpCensus,         "census");
-   VGP_(register_profile_event)(VgpCensusHeap,     "census-heap");
-   VGP_(register_profile_event)(VgpCensusSnapshot, "census-snapshot");
-   VGP_(register_profile_event)(VgpCensusTreeSize, "census-treesize");
-   VGP_(register_profile_event)(VgpUpdateXCon,     "update-XCon");
-   VGP_(register_profile_event)(VgpCalcSpacetime2, "calc-exact_ST_dbld");
-   VGP_(register_profile_event)(VgpPrintHp,        "print-hp");
-   VGP_(register_profile_event)(VgpPrintXPts,      "print-XPts");
-
-   // HP_Chunks
-   malloc_list  = VG_(HT_construct)();
-
-   // Dummy node at top of the context structure.
-   alloc_xpt = new_XPt(0, NULL, /*is_bottom*/False);
-
-   sk_assert( VG_(getcwd_alloc)(&base_dir) );
-}
-
-void SK_(post_clo_init)(void)
-{
-   ms_interval = 1;
-
-   // Do an initial sample for t = 0
-   hp_census();
-}
-
 /*------------------------------------------------------------*/
 /*--- Instrumentation                                      ---*/
 /*------------------------------------------------------------*/
 
-UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
+static IRBB* ms_instrument ( IRBB* bb_in, VexGuestLayout* layout, 
+                             IRType gWordTy, IRType hWordTy )
 {
-   return cb_in;
+   /* XXX Will Massif work when gWordTy != hWordTy ? */
+   return bb_in;
 }
 
 /*------------------------------------------------------------*/
@@ -1346,7 +1307,7 @@ static Char* clean_fnname(Char *d, Char* s)
 
 static void file_err ( Char* file )
 {
-   VG_(message)(Vg_UserMsg, "error: can't open output file `%s'", file );
+   VG_(message)(Vg_UserMsg, "error: can't open output file '%s'", file );
    VG_(message)(Vg_UserMsg, "       ... so profile results will be missing.");
 }
 
@@ -1384,12 +1345,13 @@ static void file_err ( Char* file )
  */
 static void write_hp_file(void)
 {
-   Int   i, j;
-   Int   fd, res;
-   Char *hp_file, *ps_file, *aux_file;
-   Char* cmdfmt;
-   Char* cmdbuf;
-   Int   cmdlen;
+   Int    i, j;
+   Int    fd, res;
+   SysRes sres;
+   Char  *hp_file, *ps_file, *aux_file;
+   Char*  cmdfmt;
+   Char*  cmdbuf;
+   Int    cmdlen;
 
    VGP_PUSHCC(VgpPrintHp);
    
@@ -1397,12 +1359,14 @@ static void write_hp_file(void)
    hp_file  = make_filename( base_dir, ".hp" );
    ps_file  = make_filename( base_dir, ".ps" );
    aux_file = make_filename( base_dir, ".aux" );
-   fd = VG_(open)(hp_file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
-                           VKI_S_IRUSR|VKI_S_IWUSR);
-   if (fd < 0) {
+   sres = VG_(open)(hp_file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
+                             VKI_S_IRUSR|VKI_S_IWUSR);
+   if (sres.isError) {
       file_err( hp_file );
       VGP_POPCC(VgpPrintHp);
       return;
+   } else {
+      fd = sres.val;
    }
 
    // File header, including command line
@@ -1429,10 +1393,10 @@ static void write_hp_file(void)
          for (j = 0; NULL != census->xtree_snapshots[j]; j++) {
             // Grab the jth top-XPt
             XTreeSnapshot xtree_snapshot = & census->xtree_snapshots[j][0];
-            if ( ! VG_(get_fnname)(xtree_snapshot->xpt->eip, buf2, 16)) {
+            if ( ! VG_(get_fnname)(xtree_snapshot->xpt->ip, buf2, 16)) {
                VG_(sprintf)(buf2, "???");
             }
-            SPRINTF(buf, "x%x:%s %d\n", xtree_snapshot->xpt->eip,
+            SPRINTF(buf, "x%x:%s %d\n", xtree_snapshot->xpt->ip,
                          clean_fnname(buf3, buf2), xtree_snapshot->space);
          }
 
@@ -1454,7 +1418,7 @@ static void write_hp_file(void)
    }
 
    // Close file
-   sk_assert(fd >= 0);
+   tl_assert(fd >= 0);
    VG_(close)(fd);
 
    // Attempt to convert file using hp2ps
@@ -1506,26 +1470,26 @@ static Char* make_perc(ULong spacetime, ULong total_spacetime)
    static Char mbuf[32];
    
    UInt  p = 10;
-   sk_assert(0 != total_spacetime);
+   tl_assert(0 != total_spacetime);
    percentify(spacetime * 100 * p / total_spacetime, p, 5, mbuf); 
    return mbuf;
 }
 
-// Nb: passed in XPt is a lower-level XPt;  %eips are grabbed from
+// Nb: passed in XPt is a lower-level XPt;  IPs are grabbed from
 // bottom-to-top of XCon, and then printed in the reverse order.
 static UInt pp_XCon(Int fd, XPt* xpt)
 {
-   Addr  rev_eips[clo_depth+1];
+   Addr  rev_ips[clo_depth+1];
    Int   i = 0;
    Int   n = 0;
    Bool  is_HTML      = ( XHTML == clo_format );
    Char* maybe_br     = ( is_HTML ? "<br>" : "" );
    Char* maybe_indent = ( is_HTML ? "&nbsp;&nbsp;" : "" );
 
-   sk_assert(NULL != xpt);
+   tl_assert(NULL != xpt);
 
    while (True) {
-      rev_eips[i] = xpt->eip;
+      rev_ips[i] = xpt->ip;
       n++;
       if (alloc_xpt == xpt->parent) break;
       i++;
@@ -1534,7 +1498,7 @@ static UInt pp_XCon(Int fd, XPt* xpt)
 
    for (i = n-1; i >= 0; i--) {
       // -1 means point to calling line
-      VG_(describe_eip)(rev_eips[i]-1, buf2, BUF_LEN);
+      VG_(describe_IP)(rev_ips[i]-1, buf2, BUF_LEN);
       SPRINTF(buf, "  %s%s%s\n", maybe_indent, buf2, maybe_br);
    }
 
@@ -1542,7 +1506,7 @@ static UInt pp_XCon(Int fd, XPt* xpt)
 }
 
 // Important point:  for HTML, each XPt must be identified uniquely for the
-// HTML links to all match up correctly.  Using xpt->eip is not
+// HTML links to all match up correctly.  Using xpt->ip is not
 // sufficient, because function pointers mean that you can call more than
 // one other function from a single code location.  So instead we use the
 // address of the xpt struct itself, which is guaranteed to be unique.
@@ -1557,7 +1521,7 @@ static void pp_all_XPts2(Int fd, Queue* q, ULong heap_spacetime,
    UInt  c2  = 0;
    ULong sum = 0;
    UInt  n;
-   Char *eip_desc, *perc;
+   Char *ip_desc, *perc;
    Bool  is_HTML   = ( XHTML == clo_format );
    Char* maybe_br  = ( is_HTML ?  "<br>" : "" );
    Char* maybe_p   = ( is_HTML ?   "<p>" : "" );
@@ -1579,23 +1543,23 @@ static void pp_all_XPts2(Int fd, Queue* q, ULong heap_spacetime,
 
    while (NULL != (xpt = (XPt*)dequeue(q))) {
       // Check that non-top-level XPts have a zero .approx_ST field.
-      if (xpt->parent != alloc_xpt) sk_assert( 0 == xpt->approx_ST );
+      if (xpt->parent != alloc_xpt) tl_assert( 0 == xpt->approx_ST );
 
       // Check that the sum of all children .exact_ST_dbld fields equals
       // parent's (unless alloc_xpt, when it should == 0).
       if (alloc_xpt == xpt) {
-         sk_assert(0 == xpt->exact_ST_dbld);
+         tl_assert(0 == xpt->exact_ST_dbld);
       } else {
          sum = 0;
          for (i = 0; i < xpt->n_children; i++) {
             sum += xpt->children[i]->exact_ST_dbld;
          }
-         //sk_assert(sum == xpt->exact_ST_dbld);
+         //tl_assert(sum == xpt->exact_ST_dbld);
          // It's possible that not all the children were included in the
          // exact_ST_dbld calculations.  Hopefully almost all of them were, and
          // all the important ones.
-//         sk_assert(sum <= xpt->exact_ST_dbld);
-//         sk_assert(sum * 1.05 > xpt->exact_ST_dbld );
+//         tl_assert(sum <= xpt->exact_ST_dbld);
+//         tl_assert(sum * 1.05 > xpt->exact_ST_dbld );
 //         if (sum != xpt->exact_ST_dbld) {
 //            VG_(printf)("%ld, %ld\n", sum, xpt->exact_ST_dbld);
 //         }
@@ -1618,7 +1582,7 @@ static void pp_all_XPts2(Int fd, Queue* q, ULong heap_spacetime,
                          perc);
          }
          n = pp_XCon(fd, xpt);
-         sk_assert(n == L);
+         tl_assert(n == L);
       }
 
       // Sort children by exact_ST_dbld
@@ -1641,8 +1605,8 @@ static void pp_all_XPts2(Int fd, Queue* q, ULong heap_spacetime,
          }
 
          // Remember: exact_ST_dbld is space.time *doubled*
-         perc     = make_perc(child->exact_ST_dbld / 2, total_spacetime);
-         eip_desc = VG_(describe_eip)(child->eip-1, buf2, BUF_LEN);
+         perc    = make_perc(child->exact_ST_dbld / 2, total_spacetime);
+         ip_desc = VG_(describe_IP)(child->ip-1, buf2, BUF_LEN);
          if (is_HTML) {
             SPRINTF(buf, "<li><a name=\"a%x\"></a>", child );
 
@@ -1651,9 +1615,9 @@ static void pp_all_XPts2(Int fd, Queue* q, ULong heap_spacetime,
             } else {
                SPRINTF(buf, "%s", perc);
             }
-            SPRINTF(buf, ": %s\n", eip_desc);
+            SPRINTF(buf, ": %s\n", ip_desc);
          } else {
-            SPRINTF(buf, "  %6s: %s\n\n", perc, eip_desc);
+            SPRINTF(buf, "  %6s: %s\n\n", perc, ip_desc);
          }
 
          if (child->n_children > 0) {
@@ -1697,9 +1661,10 @@ static void pp_all_XPts(Int fd, XPt* xpt, ULong heap_spacetime,
 static void
 write_text_file(ULong total_ST, ULong heap_ST)
 {
-   Int   fd, i;
-   Char* text_file;
-   Char* maybe_p = ( XHTML == clo_format ? "<p>" : "" );
+   SysRes sres;
+   Int    fd, i;
+   Char*  text_file;
+   Char*  maybe_p = ( XHTML == clo_format ? "<p>" : "" );
 
    VGP_PUSHCC(VgpPrintXPts);
 
@@ -1707,12 +1672,14 @@ write_text_file(ULong total_ST, ULong heap_ST)
    text_file = make_filename( base_dir, 
                               ( XText == clo_format ? ".txt" : ".html" ) );
 
-   fd = VG_(open)(text_file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
+   sres = VG_(open)(text_file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                              VKI_S_IRUSR|VKI_S_IWUSR);
-   if (fd < 0) {
+   if (sres.isError) {
       file_err( text_file );
       VGP_POPCC(VgpPrintXPts);
       return;
+   } else {
+      fd = sres.val;
    }
 
    // Header
@@ -1734,7 +1701,7 @@ write_text_file(ULong total_ST, ULong heap_ST)
    if (clo_heap)
       pp_all_XPts(fd, alloc_xpt, heap_ST, total_ST);
 
-   sk_assert(fd >= 0);
+   tl_assert(fd >= 0);
    VG_(close)(fd);
 
    VGP_POPCC(VgpPrintXPts);
@@ -1762,17 +1729,17 @@ print_summary(ULong total_ST, ULong heap_ST, ULong heap_admin_ST,
                    ( 0 == total_ST ? (Char*)"(n/a)"
                                    : make_perc(heap_admin_ST, total_ST) ) );
 
-   sk_assert( VG_(HT_count_nodes)(malloc_list) == n_heap_blocks );
+   tl_assert( VG_(HT_count_nodes)(malloc_list) == n_heap_blocks );
 
    // Stack(s) ----------------------------------------------------------
    if (clo_stacks) {
-      sk_assert(0 != total_ST);
       VG_(message)(Vg_UserMsg, "stack(s):          %s", 
-                   make_perc(stack_ST, total_ST) );
+                   ( 0 == stack_ST ? (Char*)"0%" 
+                                   : make_perc(stack_ST, total_ST) ) );
    }
 
    if (VG_(clo_verbosity) > 1) {
-      sk_assert(n_xpts > 0);  // always have alloc_xpt
+      tl_assert(n_xpts > 0);  // always have alloc_xpt
       VG_(message)(Vg_DebugMsg, "    allocs: %u", n_allocs);
       VG_(message)(Vg_DebugMsg, "zeroallocs: %u (%d%%)", n_zero_allocs,
                                 n_zero_allocs * 100 / n_allocs );
@@ -1792,7 +1759,7 @@ print_summary(ULong total_ST, ULong heap_ST, ULong heap_admin_ST,
    }
 }
 
-void SK_(fini)(Int exit_status)
+static void ms_fini(Int exit_status)
 {
    ULong total_ST      = 0;
    ULong heap_ST       = 0;
@@ -1810,7 +1777,74 @@ void SK_(fini)(Int exit_status)
    print_summary  ( total_ST, heap_ST, heap_admin_ST, stack_ST );
 }
 
-VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 0)
+/*------------------------------------------------------------*/
+/*--- Initialisation                                       ---*/
+/*------------------------------------------------------------*/
+
+static void ms_post_clo_init(void)
+{
+   ms_interval = 1;
+
+   // Do an initial sample for t = 0
+   hp_census();
+}
+
+static void ms_pre_clo_init()
+{ 
+   VG_(details_name)            ("Massif");
+   VG_(details_version)         (NULL);
+   VG_(details_description)     ("a space profiler");
+   VG_(details_copyright_author)("Copyright (C) 2003, Nicholas Nethercote");
+   VG_(details_bug_reports_to)  (VG_BUGS_TO);
+
+   // Basic functions
+   VG_(basic_tool_funcs)          (ms_post_clo_init,
+                                   ms_instrument,
+                                   ms_fini);
+
+   // Needs
+   VG_(needs_libc_freeres)();
+   VG_(needs_command_line_options)(ms_process_cmd_line_option,
+                                   ms_print_usage,
+                                   ms_print_debug_usage);
+   VG_(needs_client_requests)     (ms_handle_client_request);
+   VG_(needs_malloc_replacement)  (ms_malloc,
+                                   ms___builtin_new,
+                                   ms___builtin_vec_new,
+                                   ms_memalign,
+                                   ms_calloc,
+                                   ms_free,
+                                   ms___builtin_delete,
+                                   ms___builtin_vec_delete,
+                                   ms_realloc,
+                                   0 );
+
+   // Events to track
+   VG_(track_new_mem_stack_signal)( new_mem_stack_signal );
+   VG_(track_die_mem_stack_signal)( die_mem_stack_signal );
+
+   // Profiling events
+   VG_(register_profile_event)(VgpGetXPt,         "get-XPt");
+   VG_(register_profile_event)(VgpGetXPtSearch,   "get-XPt-search");
+   VG_(register_profile_event)(VgpCensus,         "census");
+   VG_(register_profile_event)(VgpCensusHeap,     "census-heap");
+   VG_(register_profile_event)(VgpCensusSnapshot, "census-snapshot");
+   VG_(register_profile_event)(VgpCensusTreeSize, "census-treesize");
+   VG_(register_profile_event)(VgpUpdateXCon,     "update-XCon");
+   VG_(register_profile_event)(VgpCalcSpacetime2, "calc-exact_ST_dbld");
+   VG_(register_profile_event)(VgpPrintHp,        "print-hp");
+   VG_(register_profile_event)(VgpPrintXPts,      "print-XPts");
+
+   // HP_Chunks
+   malloc_list  = VG_(HT_construct)( 80021 );   // prime, big
+
+   // Dummy node at top of the context structure.
+   alloc_xpt = new_XPt(0, NULL, /*is_bottom*/False);
+
+   tl_assert( VG_(getcwd)(base_dir, VKI_PATH_MAX) );
+}
+
+VG_DETERMINE_INTERFACE_VERSION(ms_pre_clo_init, 0)
 
 /*--------------------------------------------------------------------*/
 /*--- end                                                ms_main.c ---*/
