@@ -9,7 +9,7 @@
    detecting memory errors, and AddrCheck, a lightweight Valgrind tool 
    for detecting memory errors.
 
-   Copyright (C) 2000-2004 Julian Seward 
+   Copyright (C) 2000-2005 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -30,6 +30,18 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
+#include "pub_tool_basics.h"
+#include "pub_tool_errormgr.h"      // For mac_shared.h
+#include "pub_tool_execontext.h"    // For mac_shared.h
+#include "pub_tool_hashtable.h"     // For mac_shared.h
+#include "pub_tool_libcbase.h"
+#include "pub_tool_libcassert.h"
+#include "pub_tool_libcprint.h"
+#include "pub_tool_mallocfree.h"
+#include "pub_tool_options.h"
+#include "pub_tool_profile.h"       // For mac_shared.h
+#include "pub_tool_replacemalloc.h"
+#include "pub_tool_threadstate.h"
 #include "mac_shared.h"
 
 /*------------------------------------------------------------*/
@@ -41,9 +53,6 @@ static SizeT cmalloc_n_mallocs  = 0;
 static SizeT cmalloc_n_frees    = 0;
 static SizeT cmalloc_bs_mallocd = 0;
 
-/* We want a 16B redzone on heap blocks for Addrcheck and Memcheck */
-UInt VG_(vg_malloc_redzone_szB) = 16;
-
 /* Function pointers for the two tools to track interesting events. */
 void (*MAC_(new_mem_heap)) ( Addr a, SizeT len, Bool is_inited )  = NULL;
 void (*MAC_(ban_mem_heap)) ( Addr a, SizeT len )                  = NULL;
@@ -79,11 +88,11 @@ static void add_to_freed_queue ( MAC_Chunk* mc )
 
    /* Put it at the end of the freed list */
    if (freed_list_end == NULL) {
-      sk_assert(freed_list_start == NULL);
+      tl_assert(freed_list_start == NULL);
       freed_list_end    = freed_list_start = mc;
       freed_list_volume = mc->size;
    } else {
-      sk_assert(freed_list_end->next == NULL);
+      tl_assert(freed_list_end->next == NULL);
       freed_list_end->next = mc;
       freed_list_end       = mc;
       freed_list_volume += mc->size;
@@ -94,13 +103,13 @@ static void add_to_freed_queue ( MAC_Chunk* mc )
       volume below vg_clo_freelist_vol. */
 
    while (freed_list_volume > MAC_(clo_freelist_vol)) {
-      sk_assert(freed_list_start != NULL);
-      sk_assert(freed_list_end != NULL);
+      tl_assert(freed_list_start != NULL);
+      tl_assert(freed_list_end != NULL);
 
       sc1 = freed_list_start;
       freed_list_volume -= sc1->size;
       /* VG_(printf)("volume now %d\n", freed_list_volume); */
-      sk_assert(freed_list_volume >= 0);
+      tl_assert(freed_list_volume >= 0);
 
       if (freed_list_start == freed_list_end) {
          freed_list_start = freed_list_end = NULL;
@@ -132,7 +141,8 @@ MAC_Chunk* MAC_(first_matching_freed_MAC_Chunk) ( Bool (*p)(MAC_Chunk*, void*),
 
 /* Allocate its shadow chunk, put it on the appropriate list. */
 static
-void add_MAC_Chunk ( Addr p, SizeT size, MAC_AllocKind kind, VgHashTable table)
+void add_MAC_Chunk ( ThreadId tid,
+                     Addr p, SizeT size, MAC_AllocKind kind, VgHashTable table)
 {
    MAC_Chunk* mc;
 
@@ -140,7 +150,7 @@ void add_MAC_Chunk ( Addr p, SizeT size, MAC_AllocKind kind, VgHashTable table)
    mc->data      = p;
    mc->size      = size;
    mc->allockind = kind;
-   mc->where     = VG_(get_ExeContext)(VG_(get_current_or_recent_tid)());
+   mc->where     = VG_(record_ExeContext)(tid);
 
    /* Paranoia ... ensure this area is off-limits to the client, so
       the mc->data field isn't visible to the leak checker.  If memory
@@ -148,7 +158,7 @@ void add_MAC_Chunk ( Addr p, SizeT size, MAC_AllocKind kind, VgHashTable table)
       VG_(malloc) should be noaccess as far as the client is
       concerned. */
    if (!MAC_(check_noaccess)( (Addr)mc, sizeof(MAC_Chunk), NULL )) {
-      VG_(skin_panic)("add_MAC_chunk: shadow area is accessible");
+      VG_(tool_panic)("add_MAC_Chunk: shadow area is accessible");
    } 
 
    VG_(HT_add_node)( table, (VgHashNode*)mc );
@@ -182,18 +192,18 @@ static Bool complain_about_silly_args2(SizeT n, SizeT sizeB)
 
 /* Allocate memory and note change in memory available */
 __inline__
-void* MAC_(new_block) ( Addr p, SizeT size, SizeT align, UInt rzB,
+void* MAC_(new_block) ( ThreadId tid,
+                        Addr p, SizeT size, SizeT align, UInt rzB,
                         Bool is_zeroed, MAC_AllocKind kind, VgHashTable table)
 {
    VGP_PUSHCC(VgpCliMalloc);
    cmalloc_n_mallocs ++;
-   cmalloc_bs_mallocd += size;
 
    // Allocate and zero if necessary
    if (p) {
-      sk_assert(MAC_AllocCustom == kind);
+      tl_assert(MAC_AllocCustom == kind);
    } else {
-      sk_assert(MAC_AllocCustom != kind);
+      tl_assert(MAC_AllocCustom != kind);
       p = (Addr)VG_(cli_malloc)( align, size );
       if (!p) {
          VGP_POPCC(VgpCliMalloc);
@@ -202,7 +212,10 @@ void* MAC_(new_block) ( Addr p, SizeT size, SizeT align, UInt rzB,
       if (is_zeroed) VG_(memset)((void*)p, 0, size);
    }
 
-   add_MAC_Chunk( p, size, kind, table );
+   // Only update this stat if allocation succeeded.
+   cmalloc_bs_mallocd += size;
+
+   add_MAC_Chunk( tid, p, size, kind, table );
 
    MAC_(ban_mem_heap)( p-rzB, rzB );
    MAC_(new_mem_heap)( p, size, is_zeroed );
@@ -213,63 +226,64 @@ void* MAC_(new_block) ( Addr p, SizeT size, SizeT align, UInt rzB,
    return (void*)p;
 }
 
-void* SK_(malloc) ( SizeT n )
+void* MAC_(malloc) ( ThreadId tid, SizeT n )
 {
    if (complain_about_silly_args(n, "malloc")) {
       return NULL;
    } else {
-      return MAC_(new_block) ( 0, n, VG_(clo_alignment), 
-         VG_(vg_malloc_redzone_szB), /*is_zeroed*/False, MAC_AllocMalloc,
+      return MAC_(new_block) ( tid, 0, n, VG_(clo_alignment), 
+         MAC_MALLOC_REDZONE_SZB, /*is_zeroed*/False, MAC_AllocMalloc,
          MAC_(malloc_list));
    }
 }
 
-void* SK_(__builtin_new) ( SizeT n )
+void* MAC_(__builtin_new) ( ThreadId tid, SizeT n )
 {
    if (complain_about_silly_args(n, "__builtin_new")) {
       return NULL;
    } else {
-      return MAC_(new_block) ( 0, n, VG_(clo_alignment), 
-         VG_(vg_malloc_redzone_szB), /*is_zeroed*/False, MAC_AllocNew,
+      return MAC_(new_block) ( tid, 0, n, VG_(clo_alignment), 
+         MAC_MALLOC_REDZONE_SZB, /*is_zeroed*/False, MAC_AllocNew,
          MAC_(malloc_list));
    }
 }
 
-void* SK_(__builtin_vec_new) ( SizeT n )
+void* MAC_(__builtin_vec_new) ( ThreadId tid, SizeT n )
 {
    if (complain_about_silly_args(n, "__builtin_vec_new")) {
       return NULL;
    } else {
-      return MAC_(new_block) ( 0, n, VG_(clo_alignment), 
-         VG_(vg_malloc_redzone_szB), /*is_zeroed*/False, MAC_AllocNewVec,
+      return MAC_(new_block) ( tid, 0, n, VG_(clo_alignment), 
+         MAC_MALLOC_REDZONE_SZB, /*is_zeroed*/False, MAC_AllocNewVec,
          MAC_(malloc_list));
    }
 }
 
-void* SK_(memalign) ( SizeT align, SizeT n )
+void* MAC_(memalign) ( ThreadId tid, SizeT align, SizeT n )
 {
    if (complain_about_silly_args(n, "memalign")) {
       return NULL;
    } else {
-      return MAC_(new_block) ( 0, n, align, 
-         VG_(vg_malloc_redzone_szB), /*is_zeroed*/False, MAC_AllocMalloc,
+      return MAC_(new_block) ( tid, 0, n, align, 
+         MAC_MALLOC_REDZONE_SZB, /*is_zeroed*/False, MAC_AllocMalloc,
          MAC_(malloc_list));
    }
 }
 
-void* SK_(calloc) ( SizeT nmemb, SizeT size1 )
+void* MAC_(calloc) ( ThreadId tid, SizeT nmemb, SizeT size1 )
 {
    if (complain_about_silly_args2(nmemb, size1)) {
       return NULL;
    } else {
-      return MAC_(new_block) ( 0, nmemb*size1, VG_(clo_alignment),
-         VG_(vg_malloc_redzone_szB), /*is_zeroed*/True, MAC_AllocMalloc,
+      return MAC_(new_block) ( tid, 0, nmemb*size1, VG_(clo_alignment),
+         MAC_MALLOC_REDZONE_SZB, /*is_zeroed*/True, MAC_AllocMalloc,
          MAC_(malloc_list));
    }
 }
 
 static
-void die_and_free_mem ( MAC_Chunk* mc,
+void die_and_free_mem ( ThreadId tid,
+                        MAC_Chunk* mc,
                         MAC_Chunk** prev_chunks_next_ptr, SizeT rzB )
 {
    /* Note: ban redzones again -- just in case user de-banned them
@@ -287,18 +301,17 @@ void die_and_free_mem ( MAC_Chunk* mc,
    /* Put it out of harm's way for a while, if not from a client request */
    if (MAC_AllocCustom != mc->allockind) {
       /* Record where freed */
-      mc->where = VG_(get_ExeContext) ( VG_(get_current_or_recent_tid)() );
+      mc->where = VG_(record_ExeContext) ( tid );
       add_to_freed_queue ( mc );
    } else
       VG_(free) ( mc );
 }
 
 __inline__
-void MAC_(handle_free) ( Addr p, UInt rzB, MAC_AllocKind kind )
+void MAC_(handle_free) ( ThreadId tid, Addr p, UInt rzB, MAC_AllocKind kind )
 {
    MAC_Chunk*  mc;
    MAC_Chunk** prev_chunks_next_ptr;
-   ThreadId    tid = VG_(get_current_or_recent_tid)();
 
    VGP_PUSHCC(VgpCliMalloc);
 
@@ -317,31 +330,33 @@ void MAC_(handle_free) ( Addr p, UInt rzB, MAC_AllocKind kind )
       MAC_(record_freemismatch_error) ( tid, p );
    }
 
-   die_and_free_mem ( mc, prev_chunks_next_ptr, rzB );
+   die_and_free_mem ( tid, mc, prev_chunks_next_ptr, rzB );
    VGP_POPCC(VgpCliMalloc);
 }
 
-void SK_(free) ( void* p )
+void MAC_(free) ( ThreadId tid, void* p )
 {
-   MAC_(handle_free)((Addr)p, VG_(vg_malloc_redzone_szB), MAC_AllocMalloc);
+   MAC_(handle_free)( 
+      tid, (Addr)p, MAC_MALLOC_REDZONE_SZB, MAC_AllocMalloc );
 }
 
-void SK_(__builtin_delete) ( void* p )
+void MAC_(__builtin_delete) ( ThreadId tid, void* p )
 {
-   MAC_(handle_free)((Addr)p, VG_(vg_malloc_redzone_szB), MAC_AllocNew);
+   MAC_(handle_free)(
+      tid, (Addr)p, MAC_MALLOC_REDZONE_SZB, MAC_AllocNew);
 }
 
-void SK_(__builtin_vec_delete) ( void* p )
+void MAC_(__builtin_vec_delete) ( ThreadId tid, void* p )
 {
-   MAC_(handle_free)((Addr)p, VG_(vg_malloc_redzone_szB), MAC_AllocNewVec);
+   MAC_(handle_free)(
+      tid, (Addr)p, MAC_MALLOC_REDZONE_SZB, MAC_AllocNewVec);
 }
 
-void* SK_(realloc) ( void* p, SizeT new_size )
+void* MAC_(realloc) ( ThreadId tid, void* p, SizeT new_size )
 {
    MAC_Chunk  *mc;
    MAC_Chunk **prev_chunks_next_ptr;
    UInt        i;
-   ThreadId    tid = VG_(get_current_or_recent_tid)();
 
    VGP_PUSHCC(VgpCliMalloc);
 
@@ -372,7 +387,7 @@ void* SK_(realloc) ( void* p, SizeT new_size )
 
    if (mc->size == new_size) {
       /* size unchanged */
-      mc->where = VG_(get_ExeContext)(tid);
+      mc->where = VG_(record_ExeContext)(tid);
       VGP_POPCC(VgpCliMalloc);
       return p;
       
@@ -380,7 +395,7 @@ void* SK_(realloc) ( void* p, SizeT new_size )
       /* new size is smaller */
       MAC_(die_mem_heap)( mc->data+new_size, mc->size-new_size );
       mc->size = new_size;
-      mc->where = VG_(get_ExeContext)(tid);
+      mc->where = VG_(record_ExeContext)(tid);
       VGP_POPCC(VgpCliMalloc);
       return p;
 
@@ -393,24 +408,23 @@ void* SK_(realloc) ( void* p, SizeT new_size )
 
       /* First half kept and copied, second half new, 
          red zones as normal */
-      MAC_(ban_mem_heap) ( p_new-VG_(vg_malloc_redzone_szB), 
-                                 VG_(vg_malloc_redzone_szB) );
+      MAC_(ban_mem_heap) ( p_new-MAC_MALLOC_REDZONE_SZB, MAC_MALLOC_REDZONE_SZB );
       MAC_(copy_mem_heap)( (Addr)p, p_new, mc->size );
       MAC_(new_mem_heap) ( p_new+mc->size, new_size-mc->size, /*inited*/False );
-      MAC_(ban_mem_heap) ( p_new+new_size, VG_(vg_malloc_redzone_szB) );
+      MAC_(ban_mem_heap) ( p_new+new_size, MAC_MALLOC_REDZONE_SZB );
 
       /* Copy from old to new */
       for (i = 0; i < mc->size; i++)
          ((UChar*)p_new)[i] = ((UChar*)p)[i];
 
       /* Free old memory */
-      die_and_free_mem ( mc, prev_chunks_next_ptr,
-                         VG_(vg_malloc_redzone_szB) );
+      die_and_free_mem ( tid, mc, prev_chunks_next_ptr, MAC_MALLOC_REDZONE_SZB );
 
       /* this has to be after die_and_free_mem, otherwise the
          former succeeds in shorting out the new block, not the
          old, in the case when both are on the same list.  */
-      add_MAC_Chunk ( p_new, new_size, MAC_AllocMalloc,  MAC_(malloc_list) );
+      add_MAC_Chunk ( tid, p_new, new_size, 
+                           MAC_AllocMalloc, MAC_(malloc_list) );
 
       VGP_POPCC(VgpCliMalloc);
       return (void*)p_new;
@@ -427,7 +441,7 @@ void MAC_(create_mempool)(Addr pool, UInt rzB, Bool is_zeroed)
    mp->pool      = pool;
    mp->rzB       = rzB;
    mp->is_zeroed = is_zeroed;
-   mp->chunks    = VG_(HT_construct)();
+   mp->chunks    = VG_(HT_construct)( 3001 );  // prime, not so big
 
    /* Paranoia ... ensure this area is off-limits to the client, so
       the mp->data field isn't visible to the leak checker.  If memory
@@ -435,7 +449,7 @@ void MAC_(create_mempool)(Addr pool, UInt rzB, Bool is_zeroed)
       VG_(malloc) should be noaccess as far as the client is
       concerned. */
    if (!MAC_(check_noaccess)( (Addr)mp, sizeof(MAC_Mempool), NULL )) {
-      VG_(skin_panic)("MAC_(create_mempool): shadow area is accessible");
+      VG_(tool_panic)("MAC_(create_mempool): shadow area is accessible");
    } 
 
    VG_(HT_add_node)( MAC_(mempool_list), (VgHashNode*)mp );
@@ -464,7 +478,7 @@ void MAC_(destroy_mempool)(Addr pool)
                                          (void*)&prev_next );
 
    if (mp == NULL) {
-      ThreadId      tid = VG_(get_current_or_recent_tid)();
+      ThreadId tid = VG_(get_running_tid)();
 
       MAC_(record_illegal_mempool_error) ( tid, pool );
       return;
@@ -477,7 +491,7 @@ void MAC_(destroy_mempool)(Addr pool)
    VG_(free)(mp);
 }
 
-void MAC_(mempool_alloc)(Addr pool, Addr addr, SizeT size)
+void MAC_(mempool_alloc)(ThreadId tid, Addr pool, Addr addr, SizeT size)
 {
    MAC_Mempool*  mp;
    MAC_Mempool** prev_next;
@@ -486,13 +500,11 @@ void MAC_(mempool_alloc)(Addr pool, Addr addr, SizeT size)
                                         (void*)&prev_next );
 
    if (mp == NULL) {
-      ThreadId      tid = VG_(get_current_or_recent_tid)();
-
       MAC_(record_illegal_mempool_error) ( tid, pool );
       return;
    }
 
-   MAC_(new_block)(addr, size, /*ignored*/0, mp->rzB, mp->is_zeroed,
+   MAC_(new_block)(tid, addr, size, /*ignored*/0, mp->rzB, mp->is_zeroed,
                    MAC_AllocCustom, mp->chunks);
 }
 
@@ -502,8 +514,7 @@ void MAC_(mempool_free)(Addr pool, Addr addr)
    MAC_Mempool** prev_pool;
    MAC_Chunk*    mc;
    MAC_Chunk**   prev_chunk;
-   ThreadId      tid = VG_(get_current_or_recent_tid)();
-
+   ThreadId      tid = VG_(get_running_tid)();
 
    mp = (MAC_Mempool*)VG_(HT_get_node)(MAC_(mempool_list), (UWord)pool,
                                        (void*)&prev_pool);
@@ -521,9 +532,13 @@ void MAC_(mempool_free)(Addr pool, Addr addr)
       return;
    }
 
-   die_and_free_mem(mc, prev_chunk, mp->rzB);
+   die_and_free_mem ( tid, mc, prev_chunk, mp->rzB );
 }
 
+/*------------------------------------------------------------*/
+/*--- Statistics printing                                  ---*/
+/*------------------------------------------------------------*/
+
 typedef
    struct {
       UInt  nblocks;
@@ -546,9 +561,10 @@ void MAC_(print_malloc_stats) ( void )
    ms.nblocks = 0;
    ms.nbytes = 0;
    
-   /* Mmm... more lexical scoping */
    if (VG_(clo_verbosity) == 0)
       return;
+   if (VG_(clo_xml))
+      return;
 
    /* Count memory still in use. */
    VG_(HT_apply_to_all_nodes)(MAC_(malloc_list), malloc_stats_count_chunk, &ms);
@@ -565,5 +581,5 @@ void MAC_(print_malloc_stats) ( void )
 }
 
 /*--------------------------------------------------------------------*/
-/*--- end                                    mac_malloc_wrappers.c ---*/
+/*--- end                                                          ---*/
 /*--------------------------------------------------------------------*/
