@@ -7,8 +7,8 @@
    This file is part of Nulgrind, the simplest possible Valgrind tool,
    which does nothing.
 
-   Copyright (C) 2002-2004 Nicholas Nethercote
-      njn25@cam.ac.uk
+   Copyright (C) 2002-2005 Nicholas Nethercote
+      njn@valgrind.org
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -28,34 +28,41 @@
    The GNU General Public License is contained in the file COPYING.
 */
 
-#include "tool.h"
+#include "pub_tool_basics.h"
+#include "pub_tool_libcassert.h"
+#include "pub_tool_tooliface.h"
 
-void SK_(pre_clo_init)(void)
+static void nl_post_clo_init(void)
+{
+}
+
+static IRBB* nl_instrument(IRBB* bb, VexGuestLayout* layout, 
+                           IRType gWordTy, IRType hWordTy)
+{
+    return bb;
+}
+
+static void nl_fini(Int exitcode)
+{
+}
+
+static void nl_pre_clo_init(void)
 {
    VG_(details_name)            ("Nulgrind");
    VG_(details_version)         (NULL);
    VG_(details_description)     ("a binary JIT-compiler");
    VG_(details_copyright_author)(
-      "Copyright (C) 2002-2004, and GNU GPL'd, by Nicholas Nethercote.");
+      "Copyright (C) 2002-2005, and GNU GPL'd, by Nicholas Nethercote.");
    VG_(details_bug_reports_to)  (VG_BUGS_TO);
 
+   VG_(basic_tool_funcs)        (nl_post_clo_init,
+                                 nl_instrument,
+                                 nl_fini);
+
    /* No needs, no core events to track */
 }
 
-void SK_(post_clo_init)(void)
-{
-}
-
-UCodeBlock* SK_(instrument)(UCodeBlock* cb, Addr a)
-{
-    return cb;
-}
-
-void SK_(fini)(Int exitcode)
-{
-}
-
-VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 0)
+VG_DETERMINE_INTERFACE_VERSION(nl_pre_clo_init, 0)
 
 /*--------------------------------------------------------------------*/
 /*--- end                                                          ---*/
